# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"create_markdown_table": "00_utils.ipynb",
         "set_up_logging": "00_utils.ipynb",
         "request_access_token": "01_eumetsat.ipynb",
         "query_data_products": "01_eumetsat.ipynb",
         "format_dt_str": "01_eumetsat.ipynb",
         "identify_available_datasets": "01_eumetsat.ipynb",
         "dataset_id_to_link": "01_eumetsat.ipynb",
         "json_extract": "01_eumetsat.ipynb",
         "extract_metadata": "01_eumetsat.ipynb",
         "metadata_maps": "01_eumetsat.ipynb",
         "check_valid_request": "01_eumetsat.ipynb",
         "DownloadManager": "01_eumetsat.ipynb",
         "get_dir_size": "01_eumetsat.ipynb",
         "calculate_x_offset": "02_reproj.ipynb",
         "get_seviri_area_def": "02_reproj.ipynb",
         "load_scene": "02_reproj.ipynb",
         "construct_area_def": "02_reproj.ipynb",
         "construct_TM_area_def": "02_reproj.ipynb",
         "reproj_with_manual_grid": "02_reproj.ipynb",
         "reproj_to_xarray": "02_reproj.ipynb",
         "full_scene_pyresample": "02_reproj.ipynb",
         "full_scene_pyinterp": "02_reproj.ipynb",
         "Reprojector": "02_reproj.ipynb"}

modules = ["utils.py",
           "eumetsat.py",
           "reproj.py"]

doc_url = "https://AyrtonB.github.io/satip/"

git_url = "https://github.com/AyrtonB/satip/tree/master/"

def custom_doc_links(name): return None
